 /************************************************************************/
 /*                                                                      */
 /*                Centre for Speech Technology Research                 */
 /*                     University of Edinburgh, UK                      */
 /*                       Copyright (c) 1996,1997                        */
 /*                        All Rights Reserved.                          */
 /*                                                                      */
 /*  Permission to use, copy, modify, distribute this software and its   */
 /*  documentation for research, educational and individual use only, is */
 /*  hereby granted without fee, subject to the following conditions:    */
 /*   1. The code must retain the above copyright notice, this list of   */
 /*      conditions and the following disclaimer.                        */
 /*   2. Any modifications must be clearly marked as such.               */
 /*   3. Original authors' names are not deleted.                        */
 /*  This software may not be used for commercial purposes without       */
 /*  specific prior written permission from the authors.                 */
 /*                                                                      */
 /*  THE UNIVERSITY OF EDINBURGH AND THE CONTRIBUTORS TO THIS WORK       */
 /*  DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING     */
 /*  ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT  */
 /*  SHALL THE UNIVERSITY OF EDINBURGH NOR THE CONTRIBUTORS BE LIABLE    */
 /*  FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES   */
 /*  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN  */
 /*  AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,         */
 /*  ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF      */
 /*  THIS SOFTWARE.                                                      */
 /*                                                                      */
 /*************************************************************************/
 /*                                                                       */
 /*                 Author: Richard Caley (rjc@cstr.ed.ac.uk)             */
 /*                   Date: Thu Aug 14 1997                               */
 /* --------------------------------------------------------------------  */
 /* Look for subtypes of a given type defined in any module.              */
 /*                                                                       */
 /* Walks over the directory tree looking for header files with names     */
 /* which match ".*Type.h" and looks in them for class                    */
 /* definitions. Outputs definitions of subtype tables.                   */
 /*                                                                       */
 /*************************************************************************/

#include "EST.h"
#include "EST_String.h"
#include "EST_Pathname.h"
#include "EST_error.h"

#define BUFFER_SIZE (1024)

void output_type_start(EST_String base_type)
{
  
  cout << "static SubtypedObject<" << base_type << ">::Subtype " << base_type << "_subtypes[] = {\n";

}

void output_type_include(EST_String base_type)
{
  cout << "#include \"" << base_type << ".h\"\n";
}

void output_subtype_include(EST_Pathname dir, EST_String base_type, EST_String sub_type)
{
  cout << "#include \"" << EST_String(dir) + EST_String(sub_type + base_type) << ".h\"\n";
}

void output_subtype_entry(EST_String base_type, EST_String sub_type)
{
  cout << "    { \"" << sub_type << "\", " << sub_type << base_type << "::recognise, " << sub_type << base_type << "::create },\n";
}

void output_type_end(EST_String base_type)
{
  (void) base_type;
  cout << "    { NULL, NULL, NULL }\n";
  cout << "  };\n";
  cout << "const SubtypedObject<" << base_type << ">::Subtype *SubtypedObject<" << base_type << ">::subtypes = " << base_type << "_subtypes;\n";
}

void find_below(EST_Pathname dir, EST_String base_type, EST_Regex &file_re, EST_Regex &class_re, EST_StrList &dirs, EST_StrList &subtypes)
{

  bool can_have_types = access(dir + "NoSubtypes", F_OK) != 0;

  EST_StrList entries(dir.entries());

  EST_Litem *item;

  for(item=entries.head(); item; item=next(item))
    {
      EST_Pathname entry = entries(item);

      if (entry != "./" && entry != "../" && entry != "RCS/" && entry != ".\\" && entry != "RCS\\" && entry != "..\\" && entry.is_dirname())
	  find_below(dir + entry, base_type, file_re, class_re, dirs, subtypes);
      else if (can_have_types && entry.matches(file_re))
	{
	  FILE *f;
	  char buffer[BUFFER_SIZE];

	  if ((f=fopen(dir + entry, "r"))==NULL)
	    EST_sys_error("can't read '%s'", (const char *)(dir + entry));

	  while(fgets(buffer, BUFFER_SIZE, f) != NULL)
	    {
	      EST_String line(buffer);
	      int start_br[EST_Regex_max_subexpressions];
	      int end_br[EST_Regex_max_subexpressions];
	      int len;

	      if (line.search(class_re, len, 0, start_br, end_br)>=0)
		{
		  EST_String sub_type(line.at(start_br[1], end_br[1]- start_br[1]));
		  cerr << " " <<  sub_type;

		  dirs.append(dir);
		  subtypes.append(sub_type);
		}
	    }

	  fclose(f);
	}
    }
}

int main(int argc, char *argv[])
{
    if (argc <3)
	EST_error("Useage: find_subtype DIRECTORIES - TYPE..."); 

    int i;
    EST_StrList all_dirs;
    EST_Litem *d;

    for (i=1; i < argc; i++)
    {
	if (EST_String("-") == argv[i])
	    break;
	all_dirs.append(argv[i]);
    }

    argv+=i+1;
    argc-=i+1;

    cout << "\n";
    cout << "			/* This file is autogenerated by src/modules/Makefile */\n";
    cout << "#include \"festival.h\"\n";
    cout << "\n";

    while(argc > 0)
    {
	EST_String base_type(argv[0]);
	EST_String base_type_lc(downcase(base_type));
	EST_String fre = EST_String(".*\\(") + base_type + 
	    "\\|" + base_type_lc + "\\)\\.h";
	EST_Regex  file_re(fre);
	EST_String cre = EST_String("^[ \t]*class[ \t]+\\([a-zA-Z0-9_]+\\)") 
	    + base_type;
	EST_Regex  class_re(cre);

	EST_StrList subtypes, dirs;
      
	cerr << base_type << " subtypes:";

	for (d=all_dirs.head(); d != 0; d=next(d))
	    find_below(EST_Pathname(all_dirs(d)).as_directory(), 
		       base_type, file_re, class_re, dirs, subtypes);
	cerr << "\n";

	if (subtypes.length() > 0)
 	  {
	    cout << "\n";
	    output_type_include(base_type);
	    EST_Litem *st_item, *d_item;

	    for(st_item=subtypes.head(),d_item=dirs.head(); st_item; st_item=next(st_item),d_item=next(d_item))
	      output_subtype_include(dirs(d_item), base_type, subtypes(st_item));
	
	    cout << "\n";
	    output_type_start(base_type);
	    for(st_item=subtypes.head(); st_item; st_item=next(st_item))
	      output_subtype_entry(base_type, subtypes(st_item));
	    output_type_end(base_type);
	  }

	argv++;
	argc--;
    }
    return 0;
}

